Preference {
	var <>id, <>action, <type, <>label, val, <choices, <window;
	var <secondaryText, <keysDict, <guiBounds, <gHeight, <gWidth, thisView;
	
	// val is the initial value of the preference OR the index into choices
	// label is for the GUI.
	// choices will produce a drop down menu
	// action is a Function to be evaluaed after the preference is set
	// the current value and a string view are passed into the Function
	// if val is a function, a button is created
	
	// type can be \menu, \number, \trigger, \boolean
	// \menu will result in a drop down menu of choices
	// \number will simply be a value to store
	// \trigger is a button that fires an action
	// \boolean is a true / false state button
	
	// \keys are additional variables to store in a Preference instance. Stored as a Dictionary
	// can be accessed with .key and .key_ . Must be passed in as a comma seperated Associations
	
	// action is a function to run on setting a preference value. 
	// the preference object itself is passed into the function
	
	// val is an initial state for the preference
	// in a \menu Preference, this should be the index of a choice
	// in a \number Preference, it is a value
	// in a \trigger Preference, this value is ignored
	// in a \boolean, 0 is false, 1 is true
	
	*new {arg id, value, type = \menu, action, label, choices, keys;
		^super.newCopyArgs(id.asSymbol, action, type).initPreference(value, choices, label, keys);
		}

	initPreference {arg myVal, myChoices, myLabel, myKeys;
		label = myLabel.asString;
		val = myVal;
		this.choices = myChoices;
		myKeys = myKeys.asArray;
		(myKeys.size > 0).if({
			keysDict = Dictionary.new(myKeys.size);
			keysDict.addAll(myKeys);
			keysDict.keysValuesDo({arg key, val, i;
				this.addUniqueMethod(key, {arg this;
					keysDict[key];
					});
				this.addUniqueMethod((key ++ "_").asSymbol, {arg this, newval;
					keysDict[key] = newval;
					this;
					});
				})
			})
		}

	value {
		case
			{type == \boolean} { ^val == 1 }
			{type == \trigger} { ^nil }
			{true} {^val}
		}
	
	value_ {arg newValue, update = true;
		val = newValue;
		action.value(this);
		update.if({this.updateGUI});
		}
		
	curChoice {
		choices.notNil.if({
			^choices[val.max(0).min(choices.size-1)]
			}, {
			^this.value
			})
		}

	updateGUI {
		thisView.value_(val)
		}

	choices_ {arg newChoices;
		newChoices.notNil.if({
			newChoices.isKindOf(Array).not.if({
				"Choices need to be an Array".warn;
				}, {
				choices = newChoices
				});
			})
		}
	
	// sorry... counts on decorators :(
	gui {arg bounds, parent, point;
		var container;
		guiBounds = bounds ?? {Rect(0, 0, 400, 40)};
		gHeight = guiBounds.height;
		gWidth = guiBounds.width;
		parent.isNil.if({
			parent = this.makeParent;
			parent.view.decorator_(FlowLayout(Rect(0, 0, gWidth, gHeight)));
			}, {
			window = parent
			});
		this.makeContainer(parent, point);
//		parent.front;
		}
		
	makeParent {
		^window = GUI.window.new(id, guiBounds);
		}
		
	makeContainer {arg parent, point;
		point = point ?? {(gWidth * 0.2) @ (gHeight * 0.5)};
		GUI.staticText.new(parent, gWidth @ (gHeight * 0.5))
			.string_(label);
		case 
			{
			type == \number
			} {
			thisView = GUI.numberBox.new(parent, point)
			} {
			type == \boolean
			} {
			thisView = GUI.button.new(parent, point)
				.states_([
					["False", Color.black, Color(0.7, 0.3, 0.3, 0.3)],
					["True", Color.black, Color(0.3, 0.7, 0.3, 0.3)]
					])
			} {
			type == \trigger
			} {
			thisView = GUI.button.new(parent, point)
				.states_([
					["Set", Color.black, Color(0.7, 0.3, 0.3, 0.3)]
					])
			} {
			type == \menu
			} {
			thisView = GUI.popUpMenu.new(parent, point)
				.items_(
					choices.collect({arg me;
						me.asString
						})
					)
			} {
			true
			} {
			"Hmmm... something is amiss... please check the type of Preference you are creating".warn;
			};
		thisView.notNil.if({
			thisView.value_(val)
				.action_({arg me;
					this.value_(me.value)
					});
			})					
					
		}


}

PreferencePane {
	var <bounds, <title, <>prefsPath, <prefs, <window, <scrollView;
	var <files, headerH, headerW; // headers go in the window at the top
	var saveButton, hasGUI, <prefsDict;
	
	*new {arg title, bounds, prefPath;
		^super.newCopyArgs(bounds, title, prefPath).initPrefPane;
		}
	
	*open {arg prefsPath;
		^this.readArchive(prefsPath).initOpen(prefsPath);
		}
	
	// files that you may want to open or close - place in header area
	addFile {arg ... paths;
		paths.do({arg me;
			files = files.add(me)
			})
		}
		
	initOpen {arg argPrefsPath;
		prefsPath = argPrefsPath;
		}
			
	initPrefPane {
		var sBounds, sHeight, sWidth;
		hasGUI = false;
		sBounds = GUI.window.screenBounds;
		sHeight = sBounds.height;
		sWidth = sBounds.width;
		bounds = bounds ?? {Rect(100, 300, sWidth * 0.4, sHeight * 0.5 + 300)};
		prefsDict = IdentityDictionary.new;
		files = [];
		}
	
	saveAsArchive {
		this.writeTextArchive(prefsPath);
		("Preferences were saved to "++ prefsPath).postln; 
		}
		
	addPref {arg aPreference;
		prefs = prefs.add(aPreference);
		prefsDict = prefsDict.add(aPreference.id -> aPreference);
		hasGUI.if({aPreference.gui(parent: scrollView); scrollView.decorator.nextLine});
		}
	
	at {arg id;
		^prefsDict[id].value;
		}
		
	gui {
		var thisString, thisFunc;
		hasGUI = true;
		headerH = bounds.height * 0.2;
		headerW = bounds.width * 0.9; //(files.isNil.if({0.9}, {files.size.reciprocal * 0.9}));
		window = GUI.window.new(title, bounds).front;
		window.view.decorator_(FlowLayout(Rect(0, 0, bounds.height, bounds.width)));
		saveButton = GUI.button.new(window, (headerW * 0.2) @ 40)
			.states_([
				["Save Preferences", Color.black, Color.red]
				])
			.action_({
				this.saveAsArchive;
				});
		files.do({arg thisFile;
			var filename;
			filename = thisFile.basename.splitext[0];
			GUI.button.new(window, (headerW * 0.2) @ 40)
				.states_([
					["Open "++filename, Color.black, Color.red],
					])
				.action_({arg me;
					Document.open(thisFile)
					});			 
			 });
		window.view.decorator.nextLine;
		scrollView = GUI.scrollView.new(window, Rect(10, 50, bounds.width - 20, 
			bounds.height - 60));
		scrollView.decorator_(FlowLayout( (bounds.width - 20 @ bounds.height - 60).asRect));
		prefs.do({arg me;
			me.gui(parent: scrollView);
			scrollView.decorator.nextLine;
			});
		}
	

}